@@ -57,6 +57,8 @@ import Utility.Tmp.Dir
 import Utility.Env
 import Utility.Metered
 import Utility.FileMode
+import qualified Utility.RawFilePath as R
+import qualified Utility.FileIO as F
 
 import Network.URI
 import Data.Either
@@ -64,8 +66,6 @@ import Data.Char
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Char8 as B8
 import qualified Data.Map.Strict as M
-import qualified System.FilePath.ByteString as P
-import qualified Utility.RawFilePath as R
 import qualified Data.Set as S
 
 run :: [String] -> IO ()
@@ -145,13 +145,14 @@ list st rmt forpush = do
 		else downloadManifestOrFail rmt
 	l <- forM (inManifest manifest) $ \k -> do
 		b <- downloadGitBundle rmt k
-		heads <- inRepo $ Git.Bundle.listHeads b	
+		let b' = fromOsPath b
+		heads <- inRepo $ Git.Bundle.listHeads b'	
 		-- Get all the objects from the bundle. This is done here
 		-- so that the tracking refs can be updated with what is
 		-- listed, and so what when a full repush is done, all
 		-- objects are available to be pushed.
 		when forpush $
-			inRepo $ Git.Bundle.unbundle b
+			inRepo $ Git.Bundle.unbundle b'
 		-- The bundle may contain tracking refs, or regular refs,
 		-- make sure we're operating on regular refs.
 		return $ map (\(s, r) -> (fromTrackingRef rmt r, s)) heads
@@ -201,7 +202,8 @@ fetch' :: State -> Remote -> Annex ()
 fetch' st rmt = do
 	manifest <- maybe (downloadManifestOrFail rmt) pure (manifestCache st)
 	forM_ (inManifest manifest) $ \k ->
-		downloadGitBundle rmt k >>= inRepo . Git.Bundle.unbundle
+		downloadGitBundle rmt k 
+			>>= inRepo . Git.Bundle.unbundle . fromOsPath
 	-- Newline indicates end of fetch.
 	liftIO $ do
 		putStrLn ""
@@ -494,14 +496,16 @@ parseSpecialRemoteUrl url remotename = case parseURI url of
 resolveSpecialRemoteWebUrl :: String -> Annex (Maybe String)
 resolveSpecialRemoteWebUrl url
 	| "http://" `isPrefixOf` lcurl || "https://" `isPrefixOf` lcurl =
-		Url.withUrlOptionsPromptingCreds $ \uo ->
-			withTmpFile "git-remote-annex" $ \tmp h -> do
+		Url.withUrlOptionsPromptingCreds Nothing $ \uo ->
+			withTmpFile (literalOsPath "git-remote-annex") $ \tmp h -> do
 				liftIO $ hClose h
 				Url.download' nullMeterUpdate Nothing url tmp uo >>= \case
 					Left err -> giveup $ url ++ " " ++ err
 					Right () -> liftIO $
-						(headMaybe . lines)
-							<$> readFileStrict tmp
+						fmap decodeBS 
+							. headMaybe 
+							. fileLines'
+							<$> F.readFile' tmp
 	| otherwise = return Nothing
   where
 	lcurl = map toLower url
@@ -724,10 +728,10 @@ downloadManifest rmt = get mkmain >>= maybe (get mkbak) (pure . Just)
 	-- it needs to re-download it fresh every time, and the object
 	-- file should not be stored locally.
 	gettotmp dl = withOtherTmp $ \othertmp ->
-		withTmpFileIn (fromRawFilePath othertmp) "GITMANIFEST" $ \tmp tmph -> do
+		withTmpFileIn othertmp (literalOsPath "GITMANIFEST") $ \tmp tmph -> do
 			liftIO $ hClose tmph
 			_ <- dl tmp
-			b <- liftIO (B.readFile tmp)
+			b <- liftIO (F.readFile' tmp)
 			case parseManifest b of
 				Right m -> Just <$> verifyManifest rmt m
 				Left err -> giveup err
@@ -774,7 +778,7 @@ uploadManifest rmt manifest = do
 		dropKey' rmt mk
 		put mk
 
-	put mk = withTmpFile "GITMANIFEST" $ \tmp tmph -> do
+	put mk = withTmpFile (literalOsPath "GITMANIFEST") $ \tmp tmph -> do
 		liftIO $ B8.hPut tmph (formatManifest manifest)
 		liftIO $ hClose tmph
 		-- Uploading needs the key to be in the annex objects
@@ -785,13 +789,13 @@ uploadManifest rmt manifest = do
 		-- keys, which it is not.
 		objfile <- calcRepo (gitAnnexLocation mk)
 		modifyContentDir objfile $
-			linkOrCopy mk (toRawFilePath tmp) objfile Nothing >>= \case
+			linkOrCopy mk tmp objfile Nothing >>= \case
 				-- Important to set the right perms even
 				-- though the object is only present
 				-- briefly, since sending objects may rely
 				-- on or even copy file perms.
 				Just _ -> do
-					liftIO $ R.setFileMode objfile
+					liftIO $ R.setFileMode (fromOsPath objfile)
 						=<< defaultFileMode
 					freezeContent objfile
 				Nothing -> uploadfailed
@@ -839,9 +843,11 @@ parseManifest b =
  - interrupted before updating the manifest on the remote, or when a race
  - causes the uploaded manigest to be overwritten.
  -}
-lastPushedManifestFile :: UUID -> Git.Repo -> RawFilePath
-lastPushedManifestFile u r = gitAnnexDir r P.</> "git-remote-annex" 
-	P.</> fromUUID u P.</> "manifest"
+lastPushedManifestFile :: UUID -> Git.Repo -> OsPath
+lastPushedManifestFile u r = gitAnnexDir r 
+	</> literalOsPath "git-remote-annex" 
+	</> fromUUID u
+	</> literalOsPath "manifest"
 
 {- Call before uploading anything. The returned manifest has added
  - to it any bundle keys that were in the lastPushedManifestFile
@@ -857,7 +863,7 @@ startPush' rmt manifest = do
 	f <- fromRepo (lastPushedManifestFile (Remote.uuid rmt))
 	oldmanifest <- liftIO $ 
 		fromRight mempty . parseManifest
-			<$> B.readFile (fromRawFilePath f)
+			<$> F.readFile' f
 				`catchNonAsync` (const (pure mempty))
 	let oldmanifest' = mkManifest [] $
 		S.fromList (inManifest oldmanifest)
@@ -907,7 +913,7 @@ verifyManifest rmt manifest =
 --    and so more things pulled from it, etc.
 -- 3. Git bundle objects are not usually transferred between repositories
 --    except special remotes (although the user can if they want to).
-downloadGitBundle :: Remote -> Key -> Annex FilePath
+downloadGitBundle :: Remote -> Key -> Annex OsPath
 downloadGitBundle rmt k = getKeyExportLocations rmt k >>= \case
 	Nothing -> dlwith $ 
 		download rmt k (AssociatedFile Nothing) stdRetry noNotification
@@ -915,15 +921,15 @@ downloadGitBundle rmt k = getKeyExportLocations rmt k >>= \case
 		anyM getexport locs
   where
 	dlwith a = ifM a
-		( decodeBS <$> calcRepo (gitAnnexLocation k)
+		( calcRepo (gitAnnexLocation k)
 		, giveup $ "Failed to download " ++ serializeKey k
 		)
 
 	getexport loc = catchNonAsync (getexport' loc) (const (pure False))
 	getexport' loc =
-		getViaTmp rsp vc k (AssociatedFile Nothing) Nothing $ \tmp -> do
+		getViaTmp rsp vc k Nothing $ \tmp -> do
 			v <- Remote.retrieveExport (Remote.exportActions rmt)
-				k loc (decodeBS tmp) nullMeterUpdate
+				k loc tmp nullMeterUpdate
 			return (True, v)
 	rsp = Remote.retrievalSecurityPolicy rmt
 	vc = Remote.RemoteVerify rmt
@@ -948,7 +954,7 @@ checkPresentGitBundle rmt k =
 uploadGitObject :: Remote -> Key -> Annex ()
 uploadGitObject rmt k = getKeyExportLocations rmt k >>= \case
 	Just (loc:_) -> do
-		objfile <- fromRawFilePath <$> calcRepo (gitAnnexLocation k)
+		objfile <- calcRepo (gitAnnexLocation k)
 		Remote.storeExport (Remote.exportActions rmt) objfile k loc nullMeterUpdate
 	_ -> 
 		unlessM (upload rmt k (AssociatedFile Nothing) retry noNotification) $
@@ -973,14 +979,14 @@ generateGitBundle
 	-> Manifest
 	-> Annex (Key, Annex ())
 generateGitBundle rmt bs manifest =
-	withTmpFile "GITBUNDLE" $ \tmp tmph -> do
+	withTmpFile (literalOsPath "GITBUNDLE") $ \tmp tmph -> do
 		liftIO $ hClose tmph
-		inRepo $ Git.Bundle.create tmp bs
+		inRepo $ Git.Bundle.create (fromOsPath tmp) bs
 		bundlekey <- genGitBundleKey (Remote.uuid rmt)
-			(toRawFilePath tmp) nullMeterUpdate
+			tmp nullMeterUpdate
 		if (bundlekey `notElem` inManifest manifest)
 			then do
-				unlessM (moveAnnex bundlekey (AssociatedFile Nothing) (toRawFilePath tmp)) $
+				unlessM (moveAnnex bundlekey tmp) $
 					giveup "Unable to push"
 				return (bundlekey, uploadaction bundlekey)
 			else return (bundlekey, noop)
@@ -1020,7 +1026,7 @@ getKeyExportLocations rmt k = do
 keyExportLocations :: Remote -> Key -> GitConfig -> UUID -> Maybe [ExportLocation]
 keyExportLocations rmt k cfg uuid
 	| exportTree (Remote.config rmt) || importTree (Remote.config rmt) = 
-		Just $ map (\p -> mkExportLocation (".git" P.</> p)) $
+		Just $ map (\p -> mkExportLocation (literalOsPath ".git" </> p)) $
 			concatMap (`annexLocationsBare` k) cfgs
 	| otherwise = Nothing
   where
@@ -1089,7 +1095,7 @@ getRepo = getEnv "GIT_WORK_TREE" >>= \case
 	Nothing -> fixup <$> Git.CurrentRepo.get
   where
 	fixup r@(Repo { location = loc@(Local { worktree = Just _ }) }) =
-		r { location = loc { worktree = Just (P.takeDirectory (gitdir loc)) } }
+		r { location = loc { worktree = Just (takeDirectory (gitdir loc)) } }
 	fixup r = r
 
 -- Records what the git-annex branch was at the beginning of this command.
@@ -1122,11 +1128,11 @@ startAnnexBranch = ifM (null <$> Annex.Branch.siblingBranches)
 -- journal writes to a temporary directory, so that all writes
 -- to the git-annex branch by the action will be discarded.
 specialRemoteFromUrl :: StartAnnexBranch -> Annex a -> Annex a
-specialRemoteFromUrl sab a = withTmpDir "journal" $ \tmpdir -> do
+specialRemoteFromUrl sab a = withTmpDir (literalOsPath "journal") $ \tmpdir -> do
 	Annex.overrideGitConfig $ \c -> 
 		c { annexAlwaysCommit = False }
 	Annex.BranchState.changeState $ \st -> 
-		st { alternateJournal = Just (toRawFilePath tmpdir) }
+		st { alternateJournal = Just tmpdir }
 	a `finally` cleanupInitialization sab tmpdir
 
 -- If the git-annex branch did not exist when this command started,
@@ -1160,7 +1166,7 @@ specialRemoteFromUrl sab a = withTmpDir "journal" $ \tmpdir -> do
 -- involve checking out an adjusted branch. But git clone wants to do its
 -- own checkout. So no initialization is done then, and the git bundle
 -- objects are deleted.
-cleanupInitialization :: StartAnnexBranch -> FilePath -> Annex ()
+cleanupInitialization :: StartAnnexBranch -> OsPath -> Annex ()
 cleanupInitialization sab alternatejournaldir = void $ tryNonAsync $ do
 	liftIO $ mapM_ removeFile =<< dirContents alternatejournaldir
 	case sab of
@@ -1168,7 +1174,7 @@ cleanupInitialization sab alternatejournaldir = void $ tryNonAsync $ do
 		AnnexBranchCreatedEmpty r ->
 			whenM ((r ==) <$> Annex.Branch.getBranch) $ do
 				indexfile <- fromRepo gitAnnexIndex
-				liftIO $ removeWhenExistsWith R.removeLink indexfile
+				liftIO $ removeWhenExistsWith removeFile indexfile
 				-- When cloning failed and this is being
 				-- run as an exception is thrown, HEAD will
 				-- not be set to a valid value, which will
@@ -1196,9 +1202,9 @@ cleanupInitialization sab alternatejournaldir = void $ tryNonAsync $ do
 		forM_ ks $ \k -> case fromKey keyVariety k of
                 	GitBundleKey -> lockContentForRemoval k noop removeAnnex
 			_ -> noop
-		void $ liftIO $ tryIO $ removeDirectory (decodeBS annexobjectdir)
+		void $ liftIO $ tryIO $ removeDirectory annexobjectdir
 
-	notcrippledfilesystem = not <$> probeCrippledFileSystem
+	notcrippledfilesystem = not <$> isCrippledFileSystem
 
 	nonbuggygitversion = liftIO $
 		flip notElem buggygitversions <$> Git.Version.installed
