@@ -49,6 +49,7 @@ import Logs.Cluster.Basic
 import Utility.Metered
 import Utility.Env
 import Utility.Batch
+import qualified Utility.FileIO as F
 import Remote.Helper.Git
 import Remote.Helper.Messages
 import Remote.Helper.ExportImport
@@ -141,11 +142,11 @@ isGitRemoteAnnex r = "annex::" `isPrefixOf` Git.repoLocation r
  - etc.
  -}
 gitSetup :: SetupStage -> Maybe UUID -> Maybe CredPair -> RemoteConfig -> RemoteGitConfig -> Annex (RemoteConfig, UUID)
-gitSetup Init mu _ c _ = do
+gitSetup Init mu _ c gc = do
 	let location = maybe (giveup "Specify location=url") fromProposedAccepted $
 		M.lookup locationField c
 	r <- inRepo $ Git.Construct.fromRemoteLocation location False
-	r' <- tryGitConfigRead False r False
+	r' <- tryGitConfigRead gc False r False
 	let u = getUncachedUUID r'
 	if u == NoUUID
 		then giveup "git repository does not have an annex uuid"
@@ -186,10 +187,10 @@ configRead autoinit r = do
 	case (repoCheap r, annexignore, hasuuid) of
 		(_, True, _) -> return r
 		(True, _, _)
-			| remoteAnnexCheckUUID gc -> tryGitConfigRead autoinit r hasuuid
+			| remoteAnnexCheckUUID gc -> tryGitConfigRead gc autoinit r hasuuid
 			| otherwise -> return r
 		(False, _, False) -> configSpecialGitRemotes r >>= \case
-			Nothing -> tryGitConfigRead autoinit r False
+			Nothing -> tryGitConfigRead gc autoinit r False
 			Just r' -> return r'
 		_ -> return r
 
@@ -272,8 +273,8 @@ unavailable r u c gc = gen r' u c gc'
 
 {- Tries to read the config for a specified remote, updates state, and
  - returns the updated repo. -}
-tryGitConfigRead :: Bool -> Git.Repo -> Bool -> Annex Git.Repo
-tryGitConfigRead autoinit r hasuuid
+tryGitConfigRead :: RemoteGitConfig -> Bool -> Git.Repo -> Bool -> Annex Git.Repo
+tryGitConfigRead gc autoinit r hasuuid
 	| haveconfig r = return r -- already read
 	| Git.repoIsSsh r = storeUpdatedRemote $ do
 		v <- Ssh.onRemote NoConsumeStdin r
@@ -322,9 +323,9 @@ tryGitConfigRead autoinit r hasuuid
 				warning $ UnquotedString $ "Unable to parse git config from " ++ configloc
 				return $ Left exitcode
 
-	geturlconfig = Url.withUrlOptionsPromptingCreds $ \uo -> do
+	geturlconfig = Url.withUrlOptionsPromptingCreds (Just gc) $ \uo -> do
 		let url = Git.repoLocation r ++ "/config"
-		v <- withTmpFile "git-annex.tmp" $ \tmpfile h -> do
+		v <- withTmpFile (literalOsPath "git-annex.tmp") $ \tmpfile h -> do
 			liftIO $ hClose h
 			Url.download' nullMeterUpdate Nothing url tmpfile uo >>= \case
 				Right () ->
@@ -334,7 +335,7 @@ tryGitConfigRead autoinit r hasuuid
 						, Param "--null"
 						, Param "--list"
 						, Param "--file"
-						, File tmpfile
+						, File (fromOsPath tmpfile)
 						] >>= return . \case
 							Right r' -> Right r'
 							Left exitcode -> Left $ "git config exited " ++ show exitcode
@@ -439,7 +440,7 @@ inAnnex rmt st key = do
 	inAnnex' repo rmt st key
 
 inAnnex' :: Git.Repo -> Remote -> State -> Key -> Annex Bool
-inAnnex' repo rmt st@(State connpool duc _ _ _) key
+inAnnex' repo rmt st@(State connpool duc _ _ _ _) key
 	| isP2PHttp rmt = checkp2phttp
 	| Git.repoIsHttp repo = checkhttp
 	| Git.repoIsUrl repo = checkremote
@@ -448,7 +449,7 @@ inAnnex' repo rmt st@(State connpool duc _ _ _) key
 	checkp2phttp = p2pHttpClient rmt giveup (clientCheckPresent key)
 	checkhttp = do
 		gc <- Annex.getGitConfig
-		Url.withUrlOptionsPromptingCreds $ \uo -> 
+		Url.withUrlOptionsPromptingCreds (Just (gitconfig rmt)) $ \uo -> 
 			anyM (\u -> Url.checkBoth u (fromKey keySize key) uo)
 				(keyUrls gc repo rmt key)
 	checkremote = P2PHelper.checkpresent (Ssh.runProto rmt connpool (cantCheck rmt)) key
@@ -469,9 +470,9 @@ keyUrls gc repo r key = map tourl locs'
 		| remoteAnnexBare remoteconfig == Just False = annexLocationsNonBare gc key
 		| otherwise = annexLocationsBare gc key
 #ifndef mingw32_HOST_OS
-	locs' = map fromRawFilePath locs
+	locs' = map fromOsPath locs
 #else
-	locs' = map (replace "\\" "/" . fromRawFilePath) locs
+	locs' = map (replace "\\" "/" . fromOsPath) locs
 #endif
 	remoteconfig = gitconfig r
 
@@ -481,7 +482,7 @@ dropKey r st proof key = do
 	dropKey' repo r st proof key
 
 dropKey' :: Git.Repo -> Remote -> State -> Maybe SafeDropProof -> Key -> Annex ()
-dropKey' repo r st@(State connpool duc _ _ _) proof key
+dropKey' repo r st@(State connpool duc _ _ _ _) proof key
 	| isP2PHttp r = 
 		clientRemoveWithProof proof key unabletoremove r >>= \case
 			RemoveResultPlus True fanoutuuids ->
@@ -530,7 +531,7 @@ lockKey r st key callback = do
 	lockKey' repo r st key callback
 
 lockKey' :: Git.Repo -> Remote -> State -> Key -> (VerifiedCopy -> Annex r) -> Annex r
-lockKey' repo r st@(State connpool duc _ _ _) key callback
+lockKey' repo r st@(State connpool duc _ _ _ _) key callback
 	| isP2PHttp r = do	
 		showLocking r
 		p2pHttpClient r giveup (clientLockContent key) >>= \case
@@ -559,17 +560,17 @@ lockKey' repo r st@(State connpool duc _ _ _) key callback
 	failedlock = giveup "can't lock content"
 
 {- Tries to copy a key's content from a remote's annex to a file. -}
-copyFromRemote :: Remote -> State -> Key -> AssociatedFile -> FilePath -> MeterUpdate -> VerifyConfig -> Annex Verification
+copyFromRemote :: Remote -> State -> Key -> AssociatedFile -> OsPath -> MeterUpdate -> VerifyConfig -> Annex Verification
 copyFromRemote r st key file dest meterupdate vc = do
 	repo <- getRepo r
 	copyFromRemote'' repo r st key file dest meterupdate vc
 
-copyFromRemote'' :: Git.Repo -> Remote -> State -> Key -> AssociatedFile -> FilePath -> MeterUpdate -> VerifyConfig -> Annex Verification
-copyFromRemote'' repo r st@(State connpool _ _ _ _) key af dest meterupdate vc
+copyFromRemote'' :: Git.Repo -> Remote -> State -> Key -> AssociatedFile -> OsPath -> MeterUpdate -> VerifyConfig -> Annex Verification
+copyFromRemote'' repo r st@(State connpool _ _ _ _ _) key af dest meterupdate vc
 	| isP2PHttp r = copyp2phttp
 	| Git.repoIsHttp repo = verifyKeyContentIncrementally vc key $ \iv -> do
 		gc <- Annex.getGitConfig
-		ok <- Url.withUrlOptionsPromptingCreds $
+		ok <- Url.withUrlOptionsPromptingCreds (Just (gitconfig r)) $
 			Annex.Content.downloadUrl False key meterupdate iv (keyUrls gc repo r key) dest
 		unless ok $
 			giveup "failed to download content"
@@ -602,9 +603,8 @@ copyFromRemote'' repo r st@(State connpool _ _ _ _) key af dest meterupdate vc
 		<|> remoteAnnexBwLimit (gitconfig r)
 		
 	copyp2phttp = verifyKeyContentIncrementally vc key $ \iv -> do
-		startsz <- liftIO $ tryWhenExists $
-			getFileSize (toRawFilePath dest)
-		bracketIO (openBinaryFile dest ReadWriteMode) (hClose) $ \h -> do
+		startsz <- liftIO $ tryWhenExists $ getFileSize dest
+		bracketIO (F.openBinaryFile dest ReadWriteMode) (hClose) $ \h -> do
 			metered (Just meterupdate) key bwlimit $ \_ p -> do
 				p' <- case startsz of
 					Just startsz' -> liftIO $ do
@@ -616,16 +616,18 @@ copyFromRemote'' repo r st@(State connpool _ _ _ _) key af dest meterupdate vc
 					Valid -> return ()
 					Invalid -> giveup "Transfer failed"
 
-copyFromRemoteCheap :: State -> Git.Repo -> Maybe (Key -> AssociatedFile -> FilePath -> Annex ())
+copyFromRemoteCheap :: State -> Git.Repo -> Maybe (Key -> AssociatedFile -> OsPath -> Annex ())
 #ifndef mingw32_HOST_OS
 copyFromRemoteCheap st repo
 	| not $ Git.repoIsUrl repo = Just $ \key _af file -> guardUsable repo (giveup "cannot access remote") $ do
 		gc <- getGitConfigFromState st
 		loc <- liftIO $ gitAnnexLocation key repo gc
-		liftIO $ ifM (R.doesPathExist loc)
+		liftIO $ ifM (doesFileExist loc)
 			( do
 				absloc <- absPath loc
-				R.createSymbolicLink absloc (toRawFilePath file)
+				R.createSymbolicLink
+					(fromOsPath absloc)
+					(fromOsPath file)
 			, giveup "remote does not contain key"
 			)
 	| otherwise = Nothing
@@ -634,13 +636,13 @@ copyFromRemoteCheap _ _ = Nothing
 #endif
 
 {- Tries to copy a key's content to a remote's annex. -}
-copyToRemote :: Remote -> State -> Key -> AssociatedFile -> Maybe FilePath -> MeterUpdate -> Annex ()
+copyToRemote :: Remote -> State -> Key -> AssociatedFile -> Maybe OsPath -> MeterUpdate -> Annex ()
 copyToRemote r st key af o meterupdate = do
 	repo <- getRepo r
 	copyToRemote' repo r st key af o meterupdate
 
-copyToRemote' :: Git.Repo -> Remote -> State -> Key -> AssociatedFile -> Maybe FilePath -> MeterUpdate -> Annex ()
-copyToRemote' repo r st@(State connpool duc _ _ _) key af o meterupdate
+copyToRemote' :: Git.Repo -> Remote -> State -> Key -> AssociatedFile -> Maybe OsPath -> MeterUpdate -> Annex ()
+copyToRemote' repo r st@(State connpool duc _ _ _ _) key af o meterupdate
 	| isP2PHttp r = prepsendwith copyp2phttp
 	| not $ Git.repoIsUrl repo = ifM duc
 		( guardUsable repo (giveup "cannot access remote") $ commitOnCleanup repo r st $
@@ -680,9 +682,9 @@ copyToRemote' repo r st@(State connpool duc _ _ _) key af o meterupdate
 				let checksuccess = liftIO checkio >>= \case
 					Just err -> giveup err
 					Nothing -> return True
-				logStatusAfter NoLiveUpdate key $ Annex.Content.getViaTmp rsp verify key af (Just sz) $ \dest ->
+				logStatusAfter NoLiveUpdate key $ Annex.Content.getViaTmp rsp verify key (Just sz) $ \dest ->
 					metered (Just (combineMeterUpdate meterupdate p)) key bwlimit $ \_ p' -> 
-						copier object (fromRawFilePath dest) key p' checksuccess verify
+						copier object dest key p' checksuccess verify
 			)
 		unless res $
 			failedsend
@@ -718,10 +720,12 @@ fsckOnRemote r params
 		r' <- Git.Config.read r
 		environ <- getEnvironment
 		let environ' = addEntries 
-			[ ("GIT_WORK_TREE", fromRawFilePath $ Git.repoPath r')
-			, ("GIT_DIR", fromRawFilePath $ Git.localGitDir r')
+			[ ("GIT_WORK_TREE", fromOsPath $ Git.repoPath r')
+			, ("GIT_DIR", fromOsPath $ Git.localGitDir r')
 			] environ
-		batchCommandEnv program (Param "fsck" : params) (Just environ')
+		batchCommandEnv (fromOsPath program)
+			(Param "fsck" : params)
+			(Just environ')
 
 {- The passed repair action is run in the Annex monad of the remote. -}
 repairRemote :: Git.Repo -> Annex Bool -> Annex (IO Bool)
@@ -749,7 +753,7 @@ mkLocalRemoteAnnex repo = LocalRemoteAnnex repo <$> liftIO (newMVar [])
  - when possible.
  -}
 onLocal :: State -> Annex a -> Annex a
-onLocal (State _ _ _ _ lra) = onLocal' lra
+onLocal (State _ _ _ _ _ lra) = onLocal' lra
 
 onLocalRepo :: Git.Repo -> Annex a -> Annex a
 onLocalRepo repo a = do
@@ -815,7 +819,7 @@ wantHardLink = (annexHardLink <$> Annex.getGitConfig)
 	-- because they can be modified at any time.
 	<&&> (not <$> annexThin <$> Annex.getGitConfig)
 
-type FileCopier = FilePath -> FilePath -> Key -> MeterUpdate -> Annex Bool -> VerifyConfig -> Annex (Bool, Verification)
+type FileCopier = OsPath -> OsPath -> Key -> MeterUpdate -> Annex Bool -> VerifyConfig -> Annex (Bool, Verification)
 
 -- If either the remote or local repository wants to use hard links,
 -- the copier will do so (falling back to copying if a hard link cannot be
@@ -826,16 +830,16 @@ type FileCopier = FilePath -> FilePath -> Key -> MeterUpdate -> Annex Bool -> Ve
 -- done. Also returns Verified if the key's content is verified while
 -- copying it.
 mkFileCopier :: Bool -> State -> Annex FileCopier
-mkFileCopier remotewanthardlink (State _ _ copycowtried _ _) = do
+mkFileCopier remotewanthardlink (State _ _ copycowtried fastcopy _ _) = do
 	localwanthardlink <- wantHardLink
-	let linker = \src dest -> R.createLink (toRawFilePath src) (toRawFilePath dest) >> return True
+	let linker = \src dest -> R.createLink (fromOsPath src) (fromOsPath dest) >> return True
 	if remotewanthardlink || localwanthardlink
 		then return $ \src dest k p check verifyconfig ->
 			ifM (liftIO (catchBoolIO (linker src dest)))
 				( ifM check
 					( return (True, Verified)
 					, do
-						verificationOfContentFailed (toRawFilePath dest)
+						verificationOfContentFailed dest
 						return (False, UnVerified)
 					)
 				, copier src dest k p check verifyconfig
@@ -844,11 +848,11 @@ mkFileCopier remotewanthardlink (State _ _ copycowtried _ _) = do
   where
 	copier src dest k p check verifyconfig = do
 		iv <- startVerifyKeyContentIncrementally verifyconfig k
-		liftIO (fileCopier copycowtried src dest p iv) >>= \case
+		liftIO (fileCopier copycowtried fastcopy src dest p iv) >>= \case
 			Copied -> ifM check
 				( finishVerifyKeyContentIncrementally iv
 				, do
-					verificationOfContentFailed (toRawFilePath dest)
+					verificationOfContentFailed dest
 					return (False, UnVerified)
 				)
 			CopiedCoW -> unVerified check
@@ -860,24 +864,25 @@ mkFileCopier remotewanthardlink (State _ _ copycowtried _ _) = do
  - This returns False when the repository UUID is not as expected. -}
 type DeferredUUIDCheck = Annex Bool
 
-data State = State Ssh.P2PShellConnectionPool DeferredUUIDCheck CopyCoWTried (Annex (Git.Repo, GitConfig)) LocalRemoteAnnex
+data State = State Ssh.P2PShellConnectionPool DeferredUUIDCheck CopyCoWTried FastCopy (Annex (Git.Repo, GitConfig)) LocalRemoteAnnex
 
 getRepoFromState :: State -> Annex Git.Repo
-getRepoFromState (State _ _ _ a _) = fst <$> a
+getRepoFromState (State _ _ _ _ a _) = fst <$> a
 
 #ifndef mingw32_HOST_OS
 {- The config of the remote git repository, cached for speed. -}
 getGitConfigFromState :: State -> Annex GitConfig
-getGitConfigFromState (State _ _ _ a _) = snd <$> a
+getGitConfigFromState (State _ _ _ _ a _) = snd <$> a
 #endif
 
 mkState :: Git.Repo -> UUID -> RemoteGitConfig -> Annex State
 mkState r u gc = do
 	pool <- Ssh.mkP2PShellConnectionPool
 	copycowtried <- liftIO newCopyCoWTried
+	fastcopy <- getFastCopy gc
 	lra <- mkLocalRemoteAnnex r
 	(duc, getrepo) <- go
-	return $ State pool duc copycowtried getrepo lra
+	return $ State pool duc copycowtried fastcopy getrepo lra
   where
 	go
 		| remoteAnnexCheckUUID gc = return
@@ -886,7 +891,7 @@ mkState r u gc = do
 			rv <- liftIO newEmptyMVar
 			let getrepo = ifM (liftIO $ isEmptyMVar rv)
 				( do
-					r' <- tryGitConfigRead False r True
+					r' <- tryGitConfigRead gc False r True
 					let t = (r', extractGitConfig FromGitConfig r')
 					void $ liftIO $ tryPutMVar rv t
 					return t
